==> symbol-processing-aa-embeddable-2.1.0-1.0.29-sources.jar: org/jetbrains/kotlin/backend/common/actualizer/checker/IrAnnotationMatchingKmpChecker.kt <==
/*
 * Copyright 2010-2023 JetBrains s.r.o. and Kotlin Programming Language contributors.
 * Use of this source code is governed by the Apache 2.0 license that can be found in the license/LICENSE.txt file.
 */

package ksp.org.jetbrains.kotlin.backend.common.actualizer.checker

import ksp.org.jetbrains.kotlin.backend.common.actualizer.reportActualAnnotationsNotMatchExpect
import ksp.org.jetbrains.kotlin.config.LanguageFeature
import ksp.org.jetbrains.kotlin.ir.declarations.IrDeclaration
import ksp.org.jetbrains.kotlin.ir.expressions.IrConstructorCall
import ksp.org.jetbrains.kotlin.ir.symbols.IrSymbol
import ksp.org.jetbrains.kotlin.ir.symbols.IrTypeParameterSymbol
import ksp.org.jetbrains.kotlin.ir.util.fileOrNull
import ksp.org.jetbrains.kotlin.ir.util.isFakeOverride
import ksp.org.jetbrains.kotlin.resolve.calls.mpp.AbstractExpectActualAnnotationMatchChecker

internal object IrAnnotationMatchingKmpChecker : IrExpectActualChecker {

    override fun check(context: IrExpectActualChecker.Context) = with(context) {
        val languageVersionSettings = diagnosticsReporter.languageVersionSettings
        if (!languageVersionSettings.supportsFeature(LanguageFeature.MultiplatformRestrictions)) {
            return
        }

        for ((expectSymbol, actualSymbol) in expectActualMap.expectToActual) {
            if (expectSymbol is IrTypeParameterSymbol) {
                continue
            }
            val incompatibility = AbstractExpectActualAnnotationMatchChecker
                .areAnnotationsCompatible(expectSymbol, actualSymbol, containingExpectClass = null, matchingContext) ?: continue

            // If `actualSymbol` is obtained from a builtins provider (relevant only for stdlib) or via @kotlin.jvm.KotlinActual,
            // its file is null. In this case, the `expectSymbol` is used to prevent from crashing
            val reportOn = getTypealiasSymbolIfActualizedViaTypealias(expectSymbol.owner as IrDeclaration, classActualizationInfo)
                .let { it ?: getContainingActualClassIfFakeOverride(actualSymbol) }
                .let { it ?: actualSymbol }
                .takeIf { (it.owner as IrDeclaration).fileOrNull != null }
                ?: expectSymbol
            diagnosticsReporter.reportActualAnnotationsNotMatchExpect(
                incompatibility.expectSymbol as IrSymbol,
                incompatibility.actualSymbol as IrSymbol,
                incompatibility.type.mapAnnotationType { it.annotationSymbol as IrConstructorCall },
                reportOn,
            )
        }
    }

    private val IrSymbol.isFakeOverride: Boolean
        get() = (owner as IrDeclaration).isFakeOverride

    private fun getContainingActualClassIfFakeOverride(actualSymbol: IrSymbol): IrSymbol? {
        if (!actualSymbol.isFakeOverride) {
            return null
        }
        return getContainingTopLevelClass(actualSymbol.owner as IrDeclaration)?.symbol
    }
}

==> symbol-processing-aa-embeddable-2.2.0-2.0.2-sources.jar: org/jetbrains/kotlin/backend/common/actualizer/checker/IrAnnotationMatchingKmpChecker.kt <==
/*
 * Copyright 2010-2023 JetBrains s.r.o. and Kotlin Programming Language contributors.
 * Use of this source code is governed by the Apache 2.0 license that can be found in the license/LICENSE.txt file.
 */

package ksp.org.jetbrains.kotlin.backend.common.actualizer.checker

import ksp.org.jetbrains.kotlin.backend.common.actualizer.IrExpectActualMap
import ksp.org.jetbrains.kotlin.backend.common.actualizer.reportActualAnnotationsNotMatchExpect
import ksp.org.jetbrains.kotlin.config.LanguageFeature
import ksp.org.jetbrains.kotlin.ir.declarations.IrDeclaration
import ksp.org.jetbrains.kotlin.ir.expressions.IrConstructorCall
import ksp.org.jetbrains.kotlin.ir.symbols.IrSymbol
import ksp.org.jetbrains.kotlin.ir.symbols.IrTypeParameterSymbol
import ksp.org.jetbrains.kotlin.ir.util.fileOrNull
import ksp.org.jetbrains.kotlin.ir.util.isFakeOverride
import ksp.org.jetbrains.kotlin.resolve.calls.mpp.AbstractExpectActualAnnotationMatchChecker

internal object IrAnnotationMatchingKmpChecker : IrExpectActualChecker {
    @OptIn(IrExpectActualMap.MappingForCheckers::class)
    override fun check(context: IrExpectActualChecker.Context) = with(context) {
        val languageVersionSettings = diagnosticsReporter.languageVersionSettings
        if (!languageVersionSettings.supportsFeature(LanguageFeature.MultiplatformRestrictions)) {
            return
        }

        for ((expectSymbol, actualSymbol) in expectActualMap.expectToActual) {
            if (expectSymbol is IrTypeParameterSymbol) {
                continue
            }
            val incompatibility = AbstractExpectActualAnnotationMatchChecker
                .areAnnotationsCompatible(expectSymbol, actualSymbol, containingExpectClass = null, matchingContext) ?: continue

            // If `actualSymbol` is obtained from a builtins provider (relevant only for stdlib) or via @kotlin.jvm.KotlinActual,
            // its file is null. In this case, the `expectSymbol` is used to prevent from crashing
            val reportOn = getTypealiasSymbolIfActualizedViaTypealias(expectSymbol.owner as IrDeclaration, classActualizationInfo)
                .let { it ?: getContainingActualClassIfFakeOverride(actualSymbol) }
                .let { it ?: actualSymbol }
                .takeIf { (it.owner as IrDeclaration).fileOrNull != null }
                ?: expectSymbol
            diagnosticsReporter.reportActualAnnotationsNotMatchExpect(
                incompatibility.expectSymbol as IrSymbol,
                incompatibility.actualSymbol as IrSymbol,
                incompatibility.type.mapAnnotationType { it.annotationSymbol as IrConstructorCall },
                reportOn,
            )
        }
    }

    private val IrSymbol.isFakeOverride: Boolean
        get() = (owner as IrDeclaration).isFakeOverride

    private fun getContainingActualClassIfFakeOverride(actualSymbol: IrSymbol): IrSymbol? {
        if (!actualSymbol.isFakeOverride) {
            return null
        }
        return getContainingTopLevelClass(actualSymbol.owner as IrDeclaration)?.symbol
    }
}
